/**
In case you have problems with the external memory,

Check the XIo_In32 and XIo_Out32 operations to access memory.


*/

/**
 * Audio Mixer, it accepts audio from the stero line-in and the mono microphone inputs and
 * mixes the two signals and plays them on one of the line outs.
 *
 * The ac97_controller_4.0 was taken from the realtime project from 2005
 *
 */

#include "stdio.h"
#include "stdlib.h"
//#include "time.h"
#include "xbasic_types.h"
#include "xparameters.h"
#include "fsl.h"
#include "xintc.h"
#include "xuartlite.h"
#include "xuartlite_l.h"
#include "xac97.h"


#define printf xil_printf

#define ASCII_NUM_0 48
#define ASCII_NUM_1 49
#define ASCII_NUM_2 50
#define ASCII_NUM_3 51
#define ASCII_NUM_4 52
#define ASCII_RIGHT_ARROW 44	//TODO
#define ASCII_LEFT_ARROW 46		//TODO
#define MIXER_ECHO_SIGNAL_ON 666
#define MIXER_ECHO_SIGNAL_OFF 777
#define MIX_RATIO_GAIN 2

//#define SAMPLE_RATE 44100
#define SAMPLE_RATE 22000

#define INCOMING_FSL_ID 0		//codec to microblaze
#define MIX_INCOMING_FSL_ID	1	//microblaze to mixer
#define ECHO_INCOMING_FSL_ID 2
#define ECHO_OUTGOING_FSL_ID 2
#define MIX_OUTGOING_FSL_ID	1 	//mixer to microblaze
#define OUTGOING_FSL_ID 0		//microblaze to codec

#define TRUE 1
#define FALSE 0

#define MEM_BASE_ADDR 0x30000000
#define MEM_TOP_ADDR  0x300FFFFF

#define MIX_SIGNAL_COMMAND 1
#define PLAY_FROM_FILE_COMMAND 2
#define RECORD_IN_MEMORY_COMMAND 3
#define APPLY_ECHO_COMMAND 4
#define APPLY_EFFECT_2_COMMAND 5
#define EXIT_COMMAND 6

Xuint16* sound_in_mem = (Xuint16*)MEM_BASE_ADDR;
Xuint16* memory_record_stop = (Xuint16*)MEM_BASE_ADDR;

int cmd_choice = 0;
int perform_action = 0;
int mix_ratio_left = 64;
int mix_ratio_right = 64;

//structures required for ac97 init
XIntc InterruptController;
XUartLite UartController;
XAC97 codec;

//variables used to hold each sample
Xuint16 soundbyte_line_in;
Xuint16 soundbyte_mic_in;
Xuint16 soundbyte_out;

//displays the options availble to the user
void print_menu() {
	printf("%d. Mix Signals\r\n", MIX_SIGNAL_COMMAND);
	printf("%d. Play File From Memory\r\n", PLAY_FROM_FILE_COMMAND);
	printf("%d. Record File In Memory\r\n", RECORD_IN_MEMORY_COMMAND);
	printf("%d. Apply Echo\r\n", APPLY_ECHO_COMMAND);
	printf("%d. Apply Effect #2\r\n", APPLY_EFFECT_2_COMMAND);
	printf("%d. Stop\r\n\r\n", EXIT_COMMAND);
	fflush(stdout);
}

// adjust the mixing rate of signals
void adjustMixRatio(int c) {

	//increase mix of that signal
	if (c == ASCII_RIGHT_ARROW) {
		mix_ratio_right = ((mix_ratio_right + MIX_RATIO_GAIN) <= 128)? mix_ratio_right + MIX_RATIO_GAIN: mix_ratio_right;
		mix_ratio_left = ((mix_ratio_left - MIX_RATIO_GAIN) >= 0)? mix_ratio_left - MIX_RATIO_GAIN: mix_ratio_left;
	} //left arrow
	else if (c == ASCII_LEFT_ARROW) {
		mix_ratio_right = ((mix_ratio_right - MIX_RATIO_GAIN) >= 0)? mix_ratio_right - MIX_RATIO_GAIN: mix_ratio_right;
		mix_ratio_left = ((mix_ratio_left + MIX_RATIO_GAIN) <= 128)? mix_ratio_left + MIX_RATIO_GAIN: mix_ratio_left;
	}

	printf("Adjusting Ratios R:%d, L:%d\r\n", mix_ratio_right, mix_ratio_left);
}

// determine which arrow was selected and what the resulting mix sound ratio is
int isMixingRatio(int c) {	//TODO: fix it

	//right arrow
	if (c == ASCII_RIGHT_ARROW || c == ASCII_LEFT_ARROW) {
		return TRUE;
	}
	return FALSE;
}

//check if the comman being
int isCommandChoice(int c, int choice) {
	if (choice == c) { return TRUE; }
	return FALSE;
}


void loadMusicInMemory() {
	Xuint16 soundbyte;

	sound_in_mem = (Xuint16*)MEM_BASE_ADDR;
	memory_record_stop = sound_in_mem;

	printf("Recording @ %d...press a key to end\n\r", memory_record_stop );
	while( sound_in_mem < (Xuint16*)MEM_TOP_ADDR &&
	       XUartLite_mIsReceiveEmpty(XPAR_RS232_UART_1_BASEADDR) )
	{
		getfsl(soundbyte, INCOMING_FSL_ID);	//loading from *line_in*
		*sound_in_mem = soundbyte;
		sound_in_mem++;
	}

	memory_record_stop = sound_in_mem;		//to know how much was recorded
	printf("Recording Complete @ %d.\n\n\r", memory_record_stop );
}


//used to read a file from memory
void playFromFile(void) {

	Xuint16 soundbyte;
	Xuint8 keyPressed = -1;

	printf("Playback...\r\n" );
	sound_in_mem = (Xuint16*)MEM_BASE_ADDR;

	while( sound_in_mem < (Xuint16*)MEM_TOP_ADDR && sound_in_mem < memory_record_stop)
	{

		XUartLite_Recv(&UartController, &keyPressed, 1);
		soundbyte = *sound_in_mem;
		putfsl(soundbyte, OUTGOING_FSL_ID);
		sound_in_mem++;

	}

	printf("Playback Complete.\n\n\r");
}


void mixMusicSlow(void) {

	while (XUartLite_mIsReceiveEmpty(XPAR_RS232_UART_1_BASEADDR)) {

		getfsl(soundbyte_line_in, INCOMING_FSL_ID);
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

		//cputfsl(mix_ratio_left, MIX_INCOMING_FSL_ID);
		//putfsl(soundbyte_line_in, MIX_INCOMING_FSL_ID);
		//putfsl(soundbyte_mic_in, MIX_INCOMING_FSL_ID);
		//getfsl(soundbyte_out, MIX_OUTGOING_FSL_ID);

		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
		//nputfsl(soundbyte_line_in, OUTGOING_FSL_ID);
	}
}

void mixMusicEcho(void) {

	while (XUartLite_mIsReceiveEmpty(XPAR_RS232_UART_1_BASEADDR)) {

		getfsl(soundbyte_line_in, INCOMING_FSL_ID);
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

		//cputfsl(mix_ratio_left, MIX_INCOMING_FSL_ID);
		//putfsl(soundbyte_line_in, MIX_INCOMING_FSL_ID);
		//putfsl(soundbyte_mic_in, MIX_INCOMING_FSL_ID);
		//getfsl(soundbyte_out, MIX_OUTGOING_FSL_ID);

		putfsl(soundbyte_line_in, ECHO_INCOMING_FSL_ID);
		getfsl(soundbyte_out, ECHO_OUTGOING_FSL_ID);

		putfsl(soundbyte_out, OUTGOING_FSL_ID);
		putfsl(soundbyte_out, OUTGOING_FSL_ID);
	}
}


void mixMusic(void) {

	while (XUartLite_mIsReceiveEmpty(XPAR_RS232_UART_1_BASEADDR)) {

		getfsl(soundbyte_line_in, INCOMING_FSL_ID);
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

//		putfsl(mix_ratio_left, MIX_INCOMING_FSL_ID);
//		putfsl(soundbyte_line_in, MIX_INCOMING_FSL_ID);
//		putfsl(soundbyte_mic_in, MIX_INCOMING_FSL_ID);
//		getfsl(soundbyte_out, MIX_OUTGOING_FSL_ID);

		soundbyte_line_in = soundbyte_line_in*mix_ratio_left/64;
		soundbyte_mic_in = soundbyte_mic_in*mix_ratio_right/64;
		
		soundbyte_out = soundbyte_line_in + soundbyte_mic_in - soundbyte_line_in*soundbyte_mic_in/65535;

//		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
//		putfsl(soundbyte_mic_in, OUTGOING_FSL_ID);
		putfsl(soundbyte_out, OUTGOING_FSL_ID);
		putfsl(soundbyte_out, OUTGOING_FSL_ID);
	}
}


int main(void) {

	XStatus Status;
	int choice = -1;

	/********** SetUp UartLite **************************/
	Status = XUartLite_Initialize(&UartController, XPAR_RS232_UART_1_DEVICE_ID);
	if (Status != XST_SUCCESS) {
		printf("Uart initialization error\n\r\r");
    }

    XUartLite_ResetFifos(&UartController);

	Status = XUartLite_SelfTest(&UartController);
	if (Status != XST_SUCCESS) {
        printf("Uart self-test error\n\r");
    }

    /*
     * clears whatever is in the fifos
     */
    //XUartLite_ResetFifos(&UartController);	//TODO: perhaps it should be done somewhere else

	printf("\n\r\n\rApplication Started\n\r\n\r");

	/********** Set-up the Interrupts *******************/

	/*
     * Initialize the interrupt controller driver so that it's ready to use,
     * using the device ID that is generated in xparameters.h
     */
	Status = XIntc_Initialize(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID);
    if (Status != XST_SUCCESS) {
        printf("Interrupt controller initialization error\n\r");
    }

	/*
     * Connect the device driver handler that will be called when an interrupt
     * for the device occurs, the device driver handler performs the specific
     * interrupt processing for the device
     */
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID,
                      (XInterruptHandler)XUartLite_InterruptHandler, &UartController);
	if (Status != XST_SUCCESS)
    {
        printf("Interrupt controller connect error\n\r");
    }

	/*
     * Start the interrupt controller so interrupts are enabled for all
     * devices that cause interrupts. Specify real mode so that the timer
     * counter can cause interrupts through the interrupt controller.
     */
   	Status = XIntc_Start(&InterruptController, XIN_REAL_MODE);
   	if (Status != XST_SUCCESS)
   	{
        printf("Interrupt controller start error\n\r");
   	}

    /* Enable the interrupt for the timer counter and enable interrupts in
     * the microblaze processor
     */
   	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID);


	/**** initialize sound_in_mem *****/
	Status = XAC97_Initialize(&codec, XPAR_OPB_AC97_CONTROLLER_0_DEVICE_ID);
	if (Status != XST_SUCCESS)
   	{
        printf("Could not initialize AC97 Registers.\n\r");
   	}

    Status = XAC97_SelfTest(&codec);
    if (Status != XST_SUCCESS)
    {
        printf("AC97 Self-Test failed.\n\r");
    }
	init_sound(SAMPLE_RATE);

	playback_enable();
	record_enable();

	microblaze_enable_interrupts();

	print_menu();	//print the menu

    //foreground processing
	while (1) {
		Xuint8 c = -1;

		//no blocking call to check the uart buffer
		XUartLite_Recv(&UartController, &c, 1);

		//convert char to a number
		if (c != -1) {
			choice = atoi(&c);
		}

		//current state of the
		switch(cmd_choice) {
			case 0:
				// check if a selection from the main menu was made
				// only case 1 is acceptable, the rest display an error
				if (isCommandChoice(choice, MIX_SIGNAL_COMMAND) ) {
					printf("-----=== Begin Mixing ===-----\r\n");
					perform_action = MIX_SIGNAL_COMMAND;
					cmd_choice = MIX_SIGNAL_COMMAND;
					mixMusic();
				} else if (isCommandChoice(choice, PLAY_FROM_FILE_COMMAND)) {
					printf("-----=== Play From File ===-----\r\n");
					perform_action = PLAY_FROM_FILE_COMMAND;
					cmd_choice = PLAY_FROM_FILE_COMMAND;
					playFromFile();
				} else if (isCommandChoice(choice, RECORD_IN_MEMORY_COMMAND)) {
					printf("-----=== Record To Memory ===-----\r\n");
					perform_action = RECORD_IN_MEMORY_COMMAND;
					cmd_choice = RECORD_IN_MEMORY_COMMAND;
					loadMusicInMemory();
				} else {
					//printf("To start select an option from the menu.\n\r");
				}

				break;
			case MIX_SIGNAL_COMMAND:
				// 1. allow for the arrow keys to prove direction in terms of which signal should be heard louder
				// display the percentage of the mixing (start at 50%)
				// if a choice from the menu is selected implement it

				if (isCommandChoice(choice, APPLY_ECHO_COMMAND) ) {	//is echo
					printf("-----=== Add Echo ===-----\r\n");
					//perform_action = APPLY_ECHO_COMMAND;
					cmd_choice = APPLY_ECHO_COMMAND;
					mixMusicEcho();
				} else if (isCommandChoice(choice, APPLY_EFFECT_2_COMMAND)) {	//is effect
					printf("-----=== Start Effect ===-----\r\n");
					//perform_action = APPLY_EFFECT_2_COMMAND;
					cmd_choice = APPLY_EFFECT_2_COMMAND;
					mixMusicSlow();
				} else if (isCommandChoice(choice, EXIT_COMMAND)) {	//is stop
					printf("-----=== Stop Mixing ===-----\r\n");
					perform_action = 0;
					cmd_choice = 0;
					print_menu();
				} else if (isMixingRatio(c)) {	//is change in ratio
					adjustMixRatio(c);
					mixMusic();
				} else {
					mixMusic();
				}
				break;

			case PLAY_FROM_FILE_COMMAND:
				//TODO: play music from a file, does this mean that the music should be mixed in with a mic signal?

				perform_action = 0;
				cmd_choice = 0;
				print_menu();

				break;

			case RECORD_IN_MEMORY_COMMAND:
				//record the music and save it in memory, once done present the user with a choice

				perform_action = 0;
				cmd_choice = 0;
				print_menu();

				break;

			case APPLY_ECHO_COMMAND:

				//is echo then stop it
				if (isCommandChoice(choice, APPLY_ECHO_COMMAND) ) {
					printf("-----=== Stop Echo ===-----\r\n");
					//perform_action = MIX_SIGNAL_COMMAND;
					cmd_choice = MIX_SIGNAL_COMMAND;
					if (perform_action == MIX_SIGNAL_COMMAND)
						mixMusic();
					else if (perform_action == PLAY_FROM_FILE_COMMAND)
						playFromFile();
				} else if (isCommandChoice(choice, APPLY_EFFECT_2_COMMAND)) {	//is stop
					printf("-----=== Stop Echo, Start Effect ===-----\r\n");
					//perform_action = APPLY_EFFECT_2_COMMAND;
					cmd_choice = APPLY_EFFECT_2_COMMAND;
					mixMusicSlow();
				} else if (isCommandChoice(choice, EXIT_COMMAND)) {	//is stop
					printf("-----=== Stop Mixing ===-----\r\n");
					perform_action = 0;
					cmd_choice = 0;
					print_menu();
				} else if (isMixingRatio(c)) {	//is change in ratio
					adjustMixRatio(c);
					mixMusicEcho();
				} else {
					mixMusicEcho();
				}

				break;

			case APPLY_EFFECT_2_COMMAND:
				// 3. make sure that the signal mixing is already turned on
				// apply the 2nd effect
				// display the choice
				if (isCommandChoice(choice, APPLY_ECHO_COMMAND) ) {
					printf("-----=== Stop Effect, Start Echo ===-----\r\n");
					//perform_action = APPLY_ECHO_COMMAND;
					cmd_choice = APPLY_ECHO_COMMAND;
					mixMusicEcho();

				} else if (isCommandChoice(choice, APPLY_EFFECT_2_COMMAND)) {	//is stop
					printf("-----=== Stop Effect ===-----\r\n");
					//perform_action = MIX_SIGNAL_COMMAND;
					cmd_choice = MIX_SIGNAL_COMMAND;
					mixMusic();

				} else if (isCommandChoice(choice, EXIT_COMMAND)) {	//is stop
					printf("-----=== Stop Mixing ===-----\r\n");
					perform_action = 0;
					cmd_choice = 0;
					print_menu();
				} else if (isMixingRatio(c)) {	//is change in ratio
					adjustMixRatio(c);
					mixMusicSlow();
				} else {
					mixMusicSlow();
				}
				break;
			default:
				//not a valid choice
				//print to the screen?
				printf("Please select a function from the menu.\n\n\r\r");
				print_menu();
		}
	}

	playback_disable();
	record_disable();

   //disable interrupts, TODO: not sure if necessary
   //void XUartLite_DisableInterrupt(&UartController);
}



//uart waits for a command, takes in the command and then sets a command parameter in the while loop
