/**
In case you have problems with the external memory,

Check the XIo_In32 and XIo_Out32 operations to access memory.


*/

/**
 * Audio Mixer, it accepts audio from the stero line-in and the mono microphone inputs and
 * mixes the two signals and plays them on one of the line outs.
 *
 * The ac97_controller_4.0 was taken from the realtime project from 2005
 *
 */

#include "stdio.h"
#include "stdlib.h"
//#include "time.h"
#include "xbasic_types.h"
#include "xparameters.h"
#include "fsl.h"
#include "xintc.h"
#include "xuartlite.h"
#include "xuartlite_l.h"
#include "xac97.h"


#define printf xil_printf

//#define ASCII_NUM_0 48
//#define ASCII_NUM_1 49
//#define ASCII_NUM_2 50
//#define ASCII_NUM_3 51
//#define ASCII_NUM_4 52
#define ASCII_RIGHT_ARROW 44	// '<'
#define ASCII_LEFT_ARROW 46		// '>'
#define MIX_RATIO_GAIN 2.0

#define MIXER_ECHO_SIGNAL_ON 666
#define MIXER_ECHO_SIGNAL_OFF 777

//#define SAMPLE_RATE 44100
#define SAMPLE_RATE 22000

#define INCOMING_FSL_ID 0		//codec to microblaze
#define MIX_INPUT_FSL_ID 1		//microblaze to mixer
#define ECHO_INPUT_FSL_ID 2		//microblze to echo
#define ECHO_OUTPUT_FSL_ID 2	//echo to microblaze
#define MIX_OUTPUT_FSL_ID 1 	//mixer to microblaze
#define OUTGOING_FSL_ID 0		//microblaze to codec

#define TRUE 1
#define FALSE 0

#define MEM_BASE_ADDR 0x30000000
#define MEM_TOP_ADDR  0x300FFFFF

#define MIX_SIGNAL_COMMAND 1
#define PLAY_FROM_FILE_COMMAND 2
#define RECORD_IN_MEMORY_COMMAND 3
#define APPLY_ECHO_COMMAND 4
#define APPLY_EFFECT_2_COMMAND 5
#define EXIT_COMMAND 6

Xuint16* sound_in_mem = (Xuint16*)MEM_BASE_ADDR;
Xuint16* memory_record_stop = (Xuint16*)MEM_BASE_ADDR;

int cmd_choice = 0;				//one of three playing commands
int perform_action = 0;			//one of 6 commands playing + effects + exit
char message[50];				//save the messages in here
int new_message = FALSE;	//designtates if a new message is available
int record_music = FALSE;

float mix_ratio_left = 50.0;
float mix_ratio_right = 50.0;

Xuint16 echo_buffer[20];

XIntc InterruptController;
XUartLite UartController;
XAC97 codec;


void uartIntcHandler(void *CallBackRef);
void uartSendHandler(void *CallBackRef);

void print_menu() {
	printf("%d. Mix Signals\r\n", MIX_SIGNAL_COMMAND);
	printf("%d. Play File From Memory\r\n", PLAY_FROM_FILE_COMMAND);
	printf("%d. Record File In Memory\r\n", RECORD_IN_MEMORY_COMMAND);
	printf("%d. Apply Echo\r\n", APPLY_ECHO_COMMAND);
	printf("%d. Apply Effect #2\r\n", APPLY_EFFECT_2_COMMAND);
	printf("%d. Stop\r\n\r\n", EXIT_COMMAND);
	fflush(stdout);
}

// adjust the mixing rate of signals
void adjustMixRatio(int c) {

	//increase mix of that signal
	if (c == ASCII_RIGHT_ARROW) {
		mix_ratio_right = ((mix_ratio_right + MIX_RATIO_GAIN) <= 100.0)? mix_ratio_right + MIX_RATIO_GAIN: mix_ratio_right;
		mix_ratio_left = ((mix_ratio_left - MIX_RATIO_GAIN) >= 0.0)? mix_ratio_left - MIX_RATIO_GAIN: mix_ratio_left;
	} //left arrow
	else if (c == ASCII_LEFT_ARROW) {
		mix_ratio_right = ((mix_ratio_right - MIX_RATIO_GAIN) >= 0.0)? mix_ratio_right - MIX_RATIO_GAIN: mix_ratio_right;
		mix_ratio_left = ((mix_ratio_left + MIX_RATIO_GAIN) <= 100.0)? mix_ratio_left + MIX_RATIO_GAIN: mix_ratio_left;
	}

	printf("Adjusting Ratios R:%.2f, L:%.2f\r\n", mix_ratio_right, mix_ratio_left);
}

// determine which arrow was selected and what the resulting mix sound ratio is
int isMixingRatio(int c) {	//TODO: fix it

	//right arrow
	if (c == ASCII_RIGHT_ARROW || c == ASCII_LEFT_ARROW) {
		return TRUE;
	}
	return FALSE;
}

//check if the comman being
int isCommandChoice(int c, int choice) {
	if (choice == c) { return TRUE; }
	return FALSE;
}


void loadMusicInMemory() {
	Xuint16 soundbyte;

	sound_in_mem = (Xuint16*)MEM_BASE_ADDR;
	memory_record_stop = sound_in_mem;

	printf("Recording @ %d...press a key to end\n\r", memory_record_stop );

	//TODO: consider allowing sound from the mic to be mixed in
	while( sound_in_mem < (Xuint16*)MEM_TOP_ADDR && record_music )
	{
		getfsl(soundbyte, INCOMING_FSL_ID);	//loading from *line_in*
		*sound_in_mem = soundbyte;
		sound_in_mem++;
	}

	memory_record_stop = sound_in_mem;		//to know how much was recorded
	printf("Recording Complete @ %d.\n\n\r", memory_record_stop );
}


//used to read a file from memory
void playFromFile(int activateEcho, int activateEffect) {

	Xuint16 soundbyte;
	Xuint8 keyPressed = -1;

	printf("Playback...\r\n" );
	sound_in_mem = (Xuint16*)MEM_BASE_ADDR;

	//TODO: consider allowing sound from the mic to be mixed in

	while( sound_in_mem < (Xuint16*)MEM_TOP_ADDR && sound_in_mem < memory_record_stop)
	{
		soundbyte = *sound_in_mem;
		putfsl(soundbyte, OUTGOING_FSL_ID);
		sound_in_mem++;
	}

	printf("Playback Complete.\n\n\r");
}

//create a buffer, put music in buffer fetch the last piece from buffer
void mixMusicEcho() {

	Xuint16 soundbyte_line_in;
	Xuint16 soundbyte_mic_in;
	Xuint16 soundbyte_out;
	int index = 0;

	while (1) {

		getfsl(soundbyte_line_in, INCOMING_FSL_ID);	//left channel is read in first
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

		//putfsl(soundbyte_line_in, MIX_INPUT_FSL_ID);
		//getfsl(soundbyte_out, MIX_OUTPUT_FSL_ID);
		putfsl(soundbyte_line_in, ECHO_INPUT_FSL_ID);
		getfsl(soundbyte_out, ECHO_OUTPUT_FSL_ID);

		putfsl(soundbyte_out, OUTGOING_FSL_ID);

	}



}


//can only be implemented for recorded music
/*void mixMusicFForward()
{
	Xuint16 soundbyte;
	printf("high freq\n\r");
	while(1) {
		soundbyte = *sound;
		microblaze_bwrite_datafsl(soundbyte, 0);
		sound+=2;
	}
}*/

//slow-down effect
void mixMusicEffect() {

	Xuint16 soundbyte_line_in;
	Xuint16 soundbyte_mic_in;
	Xuint16 soundbyte_out;

	while (1) {

		getfsl(soundbyte_line_in, INCOMING_FSL_ID);	//left channel is read in first
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

		//mixer in here

		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
	}

}

void mixMusic() {

	Xuint16 soundbyte_line_in;
	Xuint16 soundbyte_mic_in;
	Xuint16 soundbyte_out;

	//send signals from codec to the mixer
	printf("Read from AC97 through FIFO\r\n");
	while (1) {
		getfsl(soundbyte_line_in, INCOMING_FSL_ID);	//left channel is read in first
		getfsl(soundbyte_mic_in, INCOMING_FSL_ID);

		//printf("data: %d.%d ", soundbyte_left, soundbyte_right);

		//printf("send config command to mix component, include mix ratio and start, stop\r\n");
		//putfsl((Xuint16)mix_ratio_left, MIX_INPUT_FSL_ID);   	//mix ratio right is inferred
		//TODO: should I be sending data right after the command??
		//put(soundbyte_line_in, MIX_INPUT_FSL_ID);
		//put(soundbyte_mic_in, MIX_INPUT_FSL_ID);


		//microblaze_bread_datafsl(soundbyte_out, MIXER_OUT_FSL_ID);
		//printf("Read signal from mixer\r\n");

		//receive signals from the mixer and put them in the codec
		putfsl(soundbyte_line_in, OUTGOING_FSL_ID);
	}

}


int main(void) {

	XStatus Status;
	int choice = -1, i;
	Xuint16 soundbyte;
	Xuint16 soundbyte2;


	/********** SetUp UartLite **************************/
	Status = XUartLite_Initialize(&UartController, XPAR_RS232_UART_1_DEVICE_ID);
	if (Status != XST_SUCCESS) {
		printf("Uart initialization error\n\r\r");
    }

    XUartLite_ResetFifos(&UartController);

	Status = XUartLite_SelfTest(&UartController);
	if (Status != XST_SUCCESS) {
        printf("Uart self-test error\n\r");
    }

    /*
     * clears whatever is in the fifos
     */
    XUartLite_ResetFifos(&UartController);	//TODO: perhaps it should be done somewhere else

	printf("Application Started...\n\r\n\r");

	/*
	 * Only set-up a function that listens for received interrupts.
	 */
	XUartLite_SetRecvHandler(&UartController, (XUartLite_Handler)uartIntcHandler,
                              &UartController);

	XUartLite_SetSendHandler(&UartController, (XUartLite_Handler)uartSendHandler,
                              &UartController);

	/*
	 * Enable interrupts.
	 */

	XUartLite_InterruptHandler(&UartController);

	XUartLite_EnableInterrupt(&UartController);

	if (XUartLite_mIsIntrEnabled(XPAR_RS232_UART_1_BASEADDR) == XFALSE) {
		printf("Interrupts not Ready");
	}


	/********** Set-up the Interrupts *******************/

	/*
     * Initialize the interrupt controller driver so that it's ready to use,
     * using the device ID that is generated in xparameters.h
     */
	Status = XIntc_Initialize(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID);
    if (Status != XST_SUCCESS) {
        printf("Interrupt controller initialization error\n\r");
    }

	/*
     * Connect the device driver handler that will be called when an interrupt
     * for the device occurs, the device driver handler performs the specific
     * interrupt processing for the device
     */
	Status = XIntc_Connect(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID,
                     (XInterruptHandler)XUartLite_InterruptHandler, &UartController);
	if (Status != XST_SUCCESS) {
        printf("Interrupt controller connect error\n\r");
    }

	/*
     * Start the interrupt controller so interrupts are enabled for all
     * devices that cause interrupts. Specify real mode so that the timer
     * counter can cause interrupts through the interrupt controller.
     */
   	Status = XIntc_Start(&InterruptController, XIN_REAL_MODE);
   	if (Status != XST_SUCCESS) {
        printf("Interrupt controller start error\n\r");
   	}

    /* Enable the interrupt for the timer counter and enable interrupts in
     * the microblaze processor
     */
   	XIntc_Enable(&InterruptController, XPAR_OPB_INTC_0_DEVICE_ID);

	/**** initialize sound_in_mem *****/
	Status = XAC97_Initialize(&codec, XPAR_OPB_AC97_CONTROLLER_0_DEVICE_ID);
	if (Status != XST_SUCCESS) {
        printf("Could not initialize AC97 Registers.\n\r");
   	}

    Status = XAC97_SelfTest(&codec);
    if (Status != XST_SUCCESS) {
        printf("AC97 Self-Test failed.\n\r");
    }

	for(i=0; i<20; i++) {
		echo_buffer[i] = -1;
	}

	init_sound(SAMPLE_RATE);

	playback_enable();
	record_enable();

	microblaze_enable_interrupts();

	print_menu();	//print the menu

    //foreground processing
	while (1) {

		//another way of triggering is checking if the uart buffer is full

		//current state of the

	}

	playback_disable();
	record_disable();

   //disable interrupts, TODO: not sure if necessary
   //void XUartLite_DisableInterrupt(&UartController);
}

void uartSendHandler(void *CallBackRef) {

	if (new_message) {
		XUartLite_Send(&UartController, (Xuint8 *)&message, strlen(&message));
		new_message = FALSE;
	}
}


//uart waits for a command, takes in the command and then sets a command parameter in the while loop
void uartIntcHandler(void *CallBackRef) {

	Xuint8 c = -1;
	int choice = 0;

	printf("Activated Uart Interrupt\n\r\n\r");
	//c = XUartLite_RecvByte(XPAR_RS232_UART_1_BASEADDR);
	XUartLite_Recv(&UartController, &c, 1);

	if (record_music == TRUE) {
		record_music = FALSE;
		return;
	}

	//convert char to a number
	if (c != -1) {
		choice = atoi(&c);
	} else {
		return;
	}

	//current state of the
	switch(choice) {
		case MIX_SIGNAL_COMMAND:

			if (isCommandChoice(cmd_choice, 0)) {
				strcpy(&message, (char *)"-----=== Begin Mixing ===-----\r\n\0");
				new_message = TRUE;
				perform_action = MIX_SIGNAL_COMMAND;
				cmd_choice = MIX_SIGNAL_COMMAND;
				mixMusic();
			} else {
				//printf("Inapropriate Selection.\n\r");
			}
			break;

		case PLAY_FROM_FILE_COMMAND:
			//TODO: play music from a file, does this mean that the music should be mixed in with a mic signal?
			// should I be reconfiguring the audioRegisters??
			if (isCommandChoice(cmd_choice, 0)) {
				strcpy(&message, (char *)"-----=== Play Recorded Music ===-----\r\n\0");
				new_message = TRUE;
				perform_action = PLAY_FROM_FILE_COMMAND;
				cmd_choice = PLAY_FROM_FILE_COMMAND;
				playFromFile(FALSE, FALSE);
				print_menu();
			} else {
				//printf("Inapropriate Selection.\n\r");
			}
			break;

		case RECORD_IN_MEMORY_COMMAND:

			//record the music and save it in memory, once done present the user with a choice
			if (isCommandChoice(cmd_choice, 0)) {
				strcpy(&message, (char *)"-----=== Record Music to Memory ===-----\r\n\0");
				new_message = TRUE;
				perform_action = RECORD_IN_MEMORY_COMMAND;
				cmd_choice = RECORD_IN_MEMORY_COMMAND;
				loadMusicInMemory();
				print_menu();
			} else {
				//printf("Inapropriate Selection.\n\r");
			}
			break;

		case APPLY_ECHO_COMMAND:

			// 2. make sure that signal mixing is already turned on
			// make a call for the echo to be applied
			// disply the choice

			if (isCommandChoice(cmd_choice, MIX_SIGNAL_COMMAND) &&
				isCommandChoice(perform_action, MIX_SIGNAL_COMMAND)) {

				strcpy(&message, (char *)"-----=== Apply Echo ===-----\r\n\0");
				new_message = TRUE;
				perform_action = APPLY_ECHO_COMMAND;
				cmd_choice = MIX_SIGNAL_COMMAND;
				mixMusicEcho();

			} else if (isCommandChoice(cmd_choice, MIX_SIGNAL_COMMAND)&&
				isCommandChoice(perform_action, APPLY_ECHO_COMMAND)) {

				strcpy(&message, (char *)"-----=== Stop Echo ===-----\r\n\0");
				new_message = TRUE;
				perform_action = MIX_SIGNAL_COMMAND;
				cmd_choice = MIX_SIGNAL_COMMAND;

			//NOT IMPLEMENTED RIGHT NOW
			/*} else if (isCommandChoice(cmd_choice, PLAY_FROM_FILE_COMMAND)&&
				isCommandChoice(perform_action, PLAY_FROM_FILE_COMMAND)) {


				printf("-----=== Apply Echo ===-----\r\n");
				perform_action = APPLY_ECHO_COMMAND;
				cmd_choice = PLAY_FROM_FILE_COMMAND;
			*/
			} else {
				//printf("Inapropriate Selection.\n\r");
			}
			break;

		case APPLY_EFFECT_2_COMMAND:

			if (isCommandChoice(cmd_choice, MIX_SIGNAL_COMMAND) &&
				isCommandChoice(perform_action, MIX_SIGNAL_COMMAND)) {

				strcpy(&message, (char *)"-----=== Apply Effect ===-----\r\n\0");
				new_message = TRUE;
				perform_action = APPLY_EFFECT_2_COMMAND;
				cmd_choice = MIX_SIGNAL_COMMAND;
				mixMusicEffect();

			} else if (isCommandChoice(cmd_choice, MIX_SIGNAL_COMMAND)&&
				isCommandChoice(perform_action, APPLY_EFFECT_2_COMMAND)) {

				strcpy(&message, (char *)"-----=== Stop Effect ===-----\r\n\0");
				new_message = TRUE;
				perform_action = MIX_SIGNAL_COMMAND;
				cmd_choice = MIX_SIGNAL_COMMAND;

			//NOT IMPLEMENTED RIGHT NOW
			/*} else if (isCommandChoice(cmd_choice, PLAY_FROM_FILE_COMMAND)&&
				isCommandChoice(perform_action, PLAY_FROM_FILE_COMMAND)) {


				printf("-----=== Apply Effect ===-----\r\n");
				perform_action = APPLY_EFFECT_2_COMMAND;
				cmd_choice = PLAY_FROM_FILE_COMMAND;
			*/
			} else {
				//printf("Inapropriate Selection.\n\r");
			}
			break;

		case EXIT_COMMAND:

			if (isCommandChoice(cmd_choice, MIX_SIGNAL_COMMAND) ||
				isCommandChoice(cmd_choice, PLAY_FROM_FILE_COMMAND)) {

				strcpy(&message, (char *)"Stop Music.\r\n\0");
				new_message = TRUE;
				perform_action = EXIT_COMMAND;
				cmd_choice = 0;
				print_menu();
			}
			break;

		default:

			if (isMixingRatio(c)) {	//is change in ratio
				adjustMixRatio(c);	//adjust the mixing ratio
			}

	}


}