library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library unisim;
use unisim.vcomponents.all;

ENTITY echo IS
	generic (
    	C_DWIDTH      : integer := 32
  	);
    Port (
	 	-- Global signals
      CLK : in std_logic;           -- System clock
      RESET : in std_logic;             -- System reset

		-- Slave FSL Signals
		FSL_S_Clk     : out std_logic;
		FSL_S_Read    : out std_logic;
		FSL_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
		FSL_S_Control : in  std_logic;
		FSL_S_Exists  : in  std_logic;

		-- Master FSL Signals
		FSL_M_Clk     : out std_logic;
		FSL_M_Write   : out std_logic;
		FSL_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
		FSL_M_Control : out std_logic;
		FSL_M_Full    : in  std_logic
	);
END echo;

ARCHITECTURE behavioral OF echo IS

  component SRL_FIFO is
    generic (
      C_DATA_BITS : integer;
      C_DEPTH     : integer);
    port (
      Clk             : in  std_logic;
      Reset           : in  std_logic;
      FIFO_Write      : in  std_logic;
      Data_In         : in  std_logic_vector(0 to C_DATA_BITS-1);
      FIFO_Read       : in  std_logic;
      Data_Out        : out std_logic_vector(0 to C_DATA_BITS-1);
      FIFO_Full       : out std_logic;
      FIFO_Half_Full  : out std_logic;
      FIFO_Half_Empty : out std_logic;
      Data_Exists     : out std_logic);
  end component SRL_FIFO;

  signal writeFIFO : std_logic;
  signal readFIFO : std_logic;
  signal dataOutFIFO : std_logic_vector( 0 to C_DWIDTH-1 );
  signal dataInFIFO : std_logic_vector( 0 to C_DWIDTH-1 );
  signal isFIFOFull : std_logic;
  
  signal temp :std_logic;


BEGIN
  echo_fifo : SRL_FIFO
    generic map (
       C_DATA_BITS => C_DWIDTH,
	    C_DEPTH     => 1000)
    port map (		
       Clk => CLK,
       Reset => RESET,
       FIFO_Write => writeFIFO,
       Data_In => FSL_S_DATA,
       FIFO_Read  => readFIFO,
       Data_Out  => dataOutFIFO,    
       FIFO_Full => isFIFOFull,
       Data_Exists => temp);
       
 -- reset_fifo: process (RESET)
--     variable counter : integer := 0;
--  begin
--      if RESET'event and RESET = '0' then
--          for i in 3 downto 0 loop
--            dataInFIFO <= "00000000000000000000000000000000";
--            writeFIFO <= '1';
--          end loop;
--      end if;
--  end process;   

  read_data_fsl: process (CLK, FSL_S_Exists)
    variable counter : integer := 0;
  begin
    if CLK'event and CLK = '1' then
      if FSL_S_Exists = '1' then
        FSL_S_Read <= '1';
        FSL_M_Data <= FSL_S_Data + dataOutFIFO;
        dataInFIFO <= FSL_S_Data;
        FSL_M_Write <= '1';
        if counter = 999 then
           readFIFO <= '1';
        else
           counter := counter + 1;
        end if;        
        writeFIFO <= '1';
      else
        FSL_M_Write <='0';
        FSL_S_Read <= '0';
        writeFIFO <= '0';
        readFIFO <= '0';
      end if;
    end if;
  end process;  

END behavioral;
